#include "SpiderMonkeyWrapper.h"
#include <crtdbg.h>
#include "nsCOMPtr.h"
#include "unfrozen/include/nsIDocument.h"
#include "nsIDOMHTMLDocument.h"
#include "nsIInterfaceRequestor.h"
#include "nsIServiceManager.h"
#include "unfrozen/include/nsIScriptContext.h"
#include "unfrozen/include/nsIScriptGlobalObject.h"
#include "unfrozen/include/jspubtd.h"
#include "unfrozen/include/nsIXPConnect.h"
#include <atlcomcli.h>
#include <atlsafe.h>


namespace EngineConstants{
	const int RUNTIME_SIZE = 8L * 1024L * 1024L;
	
	//Class for global object (top scope chain)
	JSClass global_class = {
		"global", JSCLASS_GLOBAL_FLAGS,
		JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
		JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub,
		JSCLASS_NO_OPTIONAL_MEMBERS
	};
}

namespace EngineHelepers{
	VARIANT ConvertJSValToVariant(JSContext *context,jsval jVal)
	{
		CComVariant vtRes;
		switch(JS_TypeOfValue(context,jVal))
		{
		case JSTYPE_NUMBER:
			int32 num;
			JS_ValueToECMAInt32(context,jVal,&num);
			vtRes = num;
			break;
		case JSTYPE_STRING:
			JSString *jsStr;

			jsStr = JS_ValueToString(context,jVal);
			if(jsStr != NULL)
				vtRes = JS_GetStringBytes(jsStr);
			break;
		case JSTYPE_BOOLEAN:
			JSBool bVal;
			JS_ValueToBoolean(context,jVal,&bVal);
			vtRes = bVal == JS_TRUE;
			break;
		case JSTYPE_OBJECT:
			JSObject *obj;
			JS_ValueToObject(context,jVal,&obj);
			if(JS_IsArrayObject(context,obj))
			{
				//if this is an array we will convert each one of it's element to variant.
				jsuint n;
				JS_GetArrayLength(context,obj,&n);
				CComSafeArray<VARIANT> vtResArr(n);
				for(jsuint i=0;i<n;i++)
				{
					jsval arrElemVal;
					JS_GetElement(context,obj,i,&arrElemVal);
					CComVariant vtArrElemVal = ConvertJSValToVariant(context,arrElemVal);
					vtResArr.SetAt(i,vtArrElemVal);
				}
				vtRes = vtResArr.Detach();
			}
			else
			{
				//creates the service for XPConeect to unwrap the element
				nsCOMPtr<nsIXPConnect> xpConnect;
				nsCOMPtr<nsIServiceManager> srvcManager;
				nsresult res = NS_GetServiceManager(getter_AddRefs(srvcManager));
				if(NS_FAILED(res))
					break;
				char xpConnectCID[NSID_LENGTH];
				nsIXPConnect::GetCID().ToProvidedString(xpConnectCID);
				res = srvcManager->GetServiceByContractID(xpConnectCID,NS_GET_IID(nsIXPConnect),getter_AddRefs(xpConnect));
				if(NS_FAILED(res))
					break;

				
				//Gets the native wrapper object.
				nsCOMPtr<nsIXPConnectWrappedNative> objectNativeWrapper;
				xpConnect->GetWrappedNativeOfJSObject(context,obj,getter_AddRefs(objectNativeWrapper));

				//Gets the wrapped HTMLElement interface
				nsCOMPtr<nsISupports> nativeObj;
				objectNativeWrapper->GetNative(getter_AddRefs(nativeObj));
				nsCOMPtr<nsIDOMHTMLElement> htmlElem = do_QueryInterface(nativeObj);
				vtRes.vt = VT_VOID | VT_BYREF;
				vtRes.byref = (void *) htmlElem;
			}
			break;
		case JSTYPE_VOID:
		case JSTYPE_NULL:
		default:
			break;
		}
		
		VARIANT res;
		::VariantInit(&res);
		vtRes.Detach(&res);
		return res;
	}

}

SpiderMonkeyWrapper::SpiderMonkeyWrapper(void)
	: m_context(NULL),
	  m_runtime(NULL),
	  m_globalObject(NULL),
	  m_destroyRuntime(false),
	  m_destroyContext(false)
{
}

SpiderMonkeyWrapper::~SpiderMonkeyWrapper(void)
{
	//clears Destroys all the compiled scripts.
	for(FilenameToCompiledScriptMapItr itr = m_fileToScript.begin();itr != m_fileToScript.end();itr++)
		JS_DestroyScript(m_context,itr->second);

	//checks if we created then we should destroy the runtime and context.
	if(m_destroyContext)
		JS_DestroyContext(m_context);
	if(m_destroyRuntime)
		JS_DestroyRuntime(m_runtime);
	
}

bool SpiderMonkeyWrapper::Init(nsIDOMHTMLElement *elem,bool seperateGlobal /* = true */,bool seperateContext /* = true */,bool seperateRuntime /* = false */)
{
	nsCOMPtr<nsIDOMWindow> pWindow = GetWindowlFromElement(elem);
	_ASSERT(pWindow != NULL);
	if(NULL == pWindow)
		return false;

	//Gets the current context.
	nsCOMPtr<nsIScriptGlobalObject> nsCurrentGlobalObject = do_QueryInterface(pWindow);
	nsCOMPtr<nsIScriptContext> nsCurrentContext = nsCurrentGlobalObject->GetContext();
	JSContext *currentContext = reinterpret_cast<JSContext*>(nsCurrentContext->GetNativeContext());
	
	//checks if we don't need to create a separate runtime then we wiil extract the window
	//from the element's window.
	if(seperateRuntime)
	{
		//creates a runtime for 8M
		m_runtime = JS_NewRuntime(EngineConstants::RUNTIME_SIZE);
		m_destroyRuntime = true;
	}
	else
		m_runtime = JS_GetRuntime(currentContext);

	_ASSERT(m_runtime != NULL);
	if(NULL == m_runtime)
		return false;

	//if we need to create a separate context we will perform 
	if(seperateContext)
	{
		m_context = JS_NewContext(m_runtime,8192);
		JS_SetOptions(m_context, JSOPTION_VAROBJFIX);
		JS_SetVersion(m_context, JSVERSION_LATEST);
		m_destroyContext = true;
	}
	else
		m_context = currentContext;

	_ASSERT(m_context != NULL);
	if(NULL == m_context)
		return false;

	//if we need to create a separate global object we need to allocate it 
	//and initialize the common JavaScript objects
	if(seperateGlobal)
	{
		m_globalObject = JS_NewObject(m_context,&EngineConstants::global_class,NULL,NULL);
		if(!JS_InitStandardClasses(m_context,m_globalObject))
			return false;
	}
	else
		m_globalObject = nsCurrentGlobalObject->GetGlobalJSObject();

	JS_SetErrorReporter(m_context, SpiderMonkeyWrapper::OnError);

	return true;
}

bool SpiderMonkeyWrapper::LoadScriptFromFile(const std::string &filename,bool execute /* = false */)
{
	//Compile the file to script , this action boosts the performance of the 
	JSScript *compiledScript = JS_CompileFile(m_context,m_globalObject,filename.data());
	_ASSERT(compiledScript != NULL);
	if(compiledScript == NULL)
		return false;
	jsval retVal;

	//Executes the script if needed.
	bool success = true;
	if(execute)
	{
		JSBool jsSuccess = JS_ExecuteScript(m_context,m_globalObject,compiledScript,&retVal);
		success &= jsSuccess == JS_TRUE;
	}
	if(success && execute)
		m_fileToScript.insert(std::make_pair(filename,compiledScript));

	return success;
}

bool SpiderMonkeyWrapper::RunScriptFunctionFromFile(const std::string &filename,const std::string &funcName,VARIANT *retVal)
{
	bool success = true;
	_ASSERT(retVal != NULL);
	
	//Gets the compiled script from file.
	FilenameToCompiledScriptMapItr fileItr = m_fileToScript.find(filename);
	if(fileItr == m_fileToScript.end())
	{
		//Could not find the compiled script meaning Load was not called 
		//calling load.
		if(!LoadScriptFromFile(filename,true))
			return false;
		//Gets the compiled script.
		fileItr = m_fileToScript.find(filename);
	}

	jsval jsRetVal;
	JSBool jsSuccess = JS_CallFunctionName(m_context,m_globalObject,funcName.data(),0,NULL,&jsRetVal);
	
	if(jsSuccess == JS_FALSE)	
		return false;

	*retVal = EngineHelepers::ConvertJSValToVariant(m_context,jsRetVal);
	return true;
}

void SpiderMonkeyWrapper::OnError(JSContext *context,const char *message,JSErrorReport *report)
{

}

nsIDOMWindow *SpiderMonkeyWrapper::GetWindowlFromElement(nsIDOMHTMLElement *elem)
{
	//Gets the document from the element since we can get the DocShell from the document.
	nsCOMPtr<nsIDOMHTMLElement> spElement = elem;
	nsCOMPtr<nsIDOMDocument> elemDoc;

	//WEFF_TODO: replace all _ASSERT with MIC_ASSERT_RETURN.
	_ASSERT(spElement != NULL);
	elem->GetOwnerDocument(getter_AddRefs(elemDoc));
	_ASSERT(elemDoc != NULL);

	//Gets the doc shell from the document.
	//WEFF_TODO - This method was copied from CMzPluginDocShellHelper::GetDocShellByDocument
	nsCOMPtr<nsIDocument> pDoc = do_QueryInterface(elemDoc);
	_ASSERT(pDoc != NULL);

	nsCOMPtr<nsISupports> docShellSupports =  pDoc->GetContainer().get();  //Get raw pointer from already add refed
	//WEFF_TODO: End of copied code CMzPluginDocShellHelper::GetDocShellByDocument.

	//Gets the window pointer from the docShell
	//WEFF_TODO - This code was copied from CMzPluginDocShellHelper::GetDOMWindow
	nsCOMPtr<nsIDOMWindow> pWindow;
	nsCOMPtr<nsIInterfaceRequestor> pIfsReq = do_QueryInterface(docShellSupports);
	_ASSERT(pIfsReq != NULL);
	pIfsReq->GetInterface(NS_GET_IID(nsIDOMWindow),(void**)getter_AddRefs(pWindow));
	return pWindow;
	//WEFF_TODO - End of copied code CMzPluginDocShellHelper::GetDOMWindow
}